__metaclass__ = type

import datetime
import logging
import time
from urlparse import urljoin
from xml.etree import cElementTree as ET

from twisted.internet import defer
from twisted.web.client import getPage

IVIEW_CONFIG = 'http://www.abc.net.au/iview/iview_config.xml'


def _parse_date(datestring):
    if not datestring:
        return None
    tm = time.strptime(datestring, '%d/%m/%Y %H:%M:%S')
    return datetime.datetime(*tm[:6])

def _parse_duration(timestring):
    if not timestring:
        return None
    timestring = timestring.replace('.', ':')
    parts = timestring.split(':')
    assert len(parts) <= 3
    duration = 0
    for value in parts:
        duration = 60 * duration + int(value)
    return duration


class Guide:
    auth_path = None
    auth_namespace = None
    channels = None
    programs = None
    series_list = None

    def __init__(self, config_url=IVIEW_CONFIG):
        self.config_url = config_url

    @defer.inlineCallbacks
    def update(self):
        logging.info("Downloading config %r", self.config_url)
        content = yield getPage(self.config_url)
        tree = ET.fromstring(content)

        params = dict((node.get('name'), node.get('value'))
                      for node in tree.findall('param'))

        self.auth_path = params['auth_path']
        self.auth_namespace = params['auth_namespace']

        # Load up programs from the channel guide.
        channel_guide_url = urljoin(self.config_url, params['xml_channels'])
        channels = yield self.load_channel_guide(channel_guide_url)
        programs = dict((program.id, program)
                        for channel in channels
                        for program in channel.programs
                        )

        # Load up the series information.
        series_list_url = urljoin(self.config_url, params['xml_series'])
        series_list = yield self.load_series_list(series_list_url, programs)

        # Load up additional series thumbnails.
        thumbnails_url = urljoin(self.config_url, params['xml_thumbnails'])
        yield self.load_thumbnails(
            thumbnails_url, params['thumbs_path'], series_list)

        self.channels = channels
        self.programs = programs
        self.series_list = series_list

    @defer.inlineCallbacks
    def load_channel_guide(self, channel_guide_url):
        logging.info("Downloading channel guide %r", channel_guide_url)
        content = yield getPage(channel_guide_url)
        tree = ET.fromstring(content)
        channels = []
        for group_node in tree.findall('group'):
            group_title = group_node.get('title')
            for channel_node in group_node.findall('channel'):
                channel = Channel(
                    shortname=channel_node.get('shortname'),
                    name=channel_node.find('name').text,
                    description=channel_node.find('description').text,
                    logo=channel_node.get('logo'),
                    group=group_title)
                channels.append(channel)
                channel_url = urljoin(
                    channel_guide_url, channel_node.get('output'))
                yield self.load_channel(channel_url, channel)
        defer.returnValue(channels)

    @defer.inlineCallbacks
    def load_channel(self, channel_url, channel):
        logging.info("Downloading channel content %r", channel_url)
        content = yield getPage(channel_url)
        tree = ET.fromstring(content)
        channel_node = tree.find('channel')
        for program_node in channel_node.findall(
            'program-title-list/program-title'):
            program = Program(
                id=int(program_node.get('id')),
                title=program_node.find('title').text,
                subtitle=program_node.find('short-title').text,
                synopsis=program_node.find('synopsis').text,
                publish_date=_parse_date(
                    program_node.find('publish-date').text),
                transmission_date=_parse_date(
                    program_node.find('transmission-date').text),
                expire_date=_parse_date(
                    program_node.find('expire-date').text),
                video_url=program_node.find('video-asset/url').text,
                duration=_parse_duration(
                    program_node.find('video-asset/duration').text))
            channel.add(program)

    @defer.inlineCallbacks
    def load_series_list(self, series_list_url, programs):
        logging.info("Downloading series list %r", series_list_url)
        content = yield getPage(series_list_url)
        tree = ET.fromstring(content)
        series_list = {}

        # Read the series URLs in parallel
        N_WORKERS = 4
        urls_to_check = [urljoin(series_list_url, node.get('publishedURL'))
                         for node in tree.findall('series-list/series')]
        @defer.inlineCallbacks
        def worker():
            while True:
                try:
                    series_url = urls_to_check.pop()
                except IndexError:
                    return
                series = yield self.load_series(series_url, programs)
                if series is not None:
                    series_list[series.id] = series
        yield defer.DeferredList([worker() for i in range(N_WORKERS)])

        defer.returnValue(series_list)

    @defer.inlineCallbacks
    def load_series(self, series_url, programs):
        logging.info("Downloading series info %r", series_url)
        content = yield getPage(series_url)
        tree = ET.fromstring(content)
        series_node = tree.find('series')
        if series_node is None:
            defer.returnValue(None)
            return

        thumbnail_node = series_node.find('image/version/url')
        if thumbnail_node is None:
            thumbnail_url = None
        else:
            thumbnail_url = thumbnail_node.text

        series = Series(
            id=int(series_node.get('id')),
            title=series_node.find('title').text,
            description=series_node.find('description').text,
            thumbnail_url=thumbnail_url)
        for program_node in series_node.findall('program-title'):
            program_id = int(program_node.get('id'))
            if program_id in programs:
                series.add(programs[program_id])
        defer.returnValue(series)

    @defer.inlineCallbacks
    def load_thumbnails(self, thumbnail_list_url, base_url, series_list):
        logging.info("Downloading thumbnail list %r", thumbnail_list_url)
        content = yield getPage(thumbnail_list_url)
        tree = ET.fromstring(content)
        for series_node in tree.findall('series-thumbs/item'):
            series_id = int(series_node.get('id'))
            if series_id not in series_list:
                continue
            series = series_list[series_id]
            thumbnail_url = urljoin(
                base_url, series_node.find('thumb').text)
            if series.thumbnail_url is None:
                series.thumbnail_url = thumbnail_url


    @defer.inlineCallbacks
    def get_auth_info(self):
        """Return information needed to authenticate to the RTMP server.

        Returns a (rtmp_url, auth, tcurl) tuple.
        """
        logging.info("Downloading auth URL %r", self.auth_path)
        content = yield getPage(self.auth_path)
        tree = ET.fromstring(content)
        server = tree.find("{%s}server" % self.auth_namespace).text
        token = tree.find("{%s}token" % self.auth_namespace).text
        if '&' in token:
            token = token.split('&', 1)[0]
        defer.returnValue((server, token))


class Channel:
    def __init__(self, shortname, name, description, logo, group):
        self.shortname = shortname
        self.name = name
        self.description = description
        self.logo = logo
        self.group = group
        self.programs = []

    def add(self, program):
        self.programs.append(program)

    def __repr__(self):
        return '<Channel %s>' % self.shortname


class Program:
    def __init__(self, id, title, subtitle, synopsis, publish_date,
                 expire_date, transmission_date, video_url, duration):
        self.id = id
        self.title = title
        self.subtitle = subtitle
        self.synopsis = synopsis
        self.publish_date = publish_date
        self.expire_date = expire_date
        self.transmission_date = transmission_date
        self.video_url = video_url
        self.duration = duration
        self.series = None


class Series:
    def __init__(self, id, title, description, thumbnail_url):
        self.id = id
        self.title = title
        self.description = description
        self.thumbnail_url = thumbnail_url
        self.programs = []

    def add(self, program):
        self.programs.append(program)
        program.series = self
