import dbus
import dbus.service
from twisted.internet import defer

from .dbusutil import dbus_deferred_method

MEDIA_CONTAINER_IFACE = 'org.gnome.UPnP.MediaContainer1'
MEDIA_ITEM_IFACE = 'org.gnome.UPnP.MediaItem1'
MEDIA_OBJECT_IFACE = 'org.gnome.UPnP.MediaObject1'


class MediaObjectMixin(object):

    def get_properties(self, iface):
        if iface == MEDIA_OBJECT_IFACE:
            return {
                'DisplayName': self.display_name,
                'Parent': self.parent,
                }
        else:
            return {}

    @dbus_deferred_method(dbus.PROPERTIES_IFACE,
                          in_signature='ss', out_signature='v')
    def Get(self, iface, name):
        d = defer.maybeDeferred(self.get_properties, iface)
        def get_value(dict):
            return dict[name]
        d.addCallback(get_value)
        return d

    @dbus_deferred_method(dbus.PROPERTIES_IFACE,
                          in_signature='s', out_signature='a{sv}')
    def GetAll(self, iface):
        d = defer.maybeDeferred(self.get_properties, iface)
        def cast_dict(result):
            return dbus.Dictionary(result, signature='sv')
        d.addCallback(cast_dict)
        return d


class MediaContainerMixin(MediaObjectMixin):

    def get_children(self):
        """Return lists of child items and containers."""
        raise NotImplementedError

    def get_properties(self, iface):
        if iface == MEDIA_CONTAINER_IFACE:
            d = defer.maybeDeferred(self.get_children)
            def make_properties((items, containers)):
                items = dbus.Array(items, signature='o')
                containers = dbus.Array(containers, signature='o')
                return {
                    'Items': items,
                    'ItemsCount': len(items),
                    'Containers': containers,
                    'ContainersCount': len(containers),
                    }
            d.addCallback(make_properties)
            return d
        else:
            return super(MediaContainerMixin, self).get_properties(iface)

    @dbus.service.signal(MEDIA_CONTAINER_IFACE, signature='')
    def Updated(self):
        pass



class MediaItemMixin(MediaObjectMixin):

    def get_details(self):
        """Return item details"""
        raise NotImplementedError

    def get_properties(self, iface):
        if iface == MEDIA_ITEM_IFACE:
            return self.get_details()
        else:
            return super(MediaItemMixin, self).get_properties(iface)


class IViewMediaServer(dbus.service.Object, MediaContainerMixin):
    def __init__(self, bus, appname, guide, publisher_url):
        self.path = '/org/gnome/UPnP/MediaServer1/%s' % appname
        super(IViewMediaServer, self).__init__(bus, object_path=self.path)
        self.bus = bus
        self.bus_name = dbus.service.BusName(
            'org.gnome.UPnP.MediaServer1.%s' % appname, bus)
        self.display_name = 'ABC iView'

        self.channels = []

        self.guide = guide
        self.publisher_url = publisher_url
        d = self.guide.update()
        d.addCallback(self.update)

    def update(self, result):
        existing_channels = dict(
            (channel.name, channel) for channel in self.channels)
        new_channels = []
        needs_update = False
        for guide_channel in self.guide.channels:
            channel = existing_channels.pop(guide_channel.shortname, None)
            if channel is None:
                needs_update = True
                channel = IViewChannel(self, guide_channel)
            else:
                channel.update(guide_channel)
            new_channels.append(channel)
        # Remove old channels.
        for channel in existing_channels.values():
            channel.remove_from_connection()
            needs_update = True
        self.channels = new_channels
        self.Updated()
        return result

    @property
    def parent(self):
        return self

    def get_children(self):
        return [], self.channels


class IViewChannel(dbus.service.Object, MediaContainerMixin):
    def __init__(self, parent, channel):
        self.path = '%s/channels/%s' % (parent.path, channel.shortname)
        super(IViewChannel, self).__init__(parent.bus, object_path=self.path)
        self.bus = parent.bus
        self.name = channel.shortname
        self.parent = parent
        self.display_name = None
        self.programs = []
        self.update(channel)

    def update(self, channel):
        self.display_name = channel.name
        existing_programs = dict(
            (program.id, program) for program in self.programs)
        new_programs = []
        for guide_program in reversed(channel.programs):
            program = existing_programs.pop(guide_program.id, None)
            if program is None:
                program = IViewProgram(self, guide_program)
            else:
                program.update(guide_program)
            new_programs.append(program)
        # Remove old programs
        for program in existing_programs:
            program.remove_from_connection()
        self.programs = new_programs
        self.Updated()

    def get_children(self):
        return self.programs, []


class IViewProgram(dbus.service.Object, MediaItemMixin):
    def __init__(self, parent, program):
        self.path = '%s/%d' % (parent.path, program.id)
        super(IViewProgram, self).__init__(parent.bus, object_path=self.path)
        self.bus = parent.bus
        self.id = program.id
        self.parent = parent
        self.display_name = None
        self.details = {}
        self.update(program)

    def update(self, program):
        self.display_name = program.title
        url = '%s/program/%d' % (self.parent.parent.publisher_url, program.id)
        self.details = {
            'URLs': dbus.Array([url], signature='s'),
            'MIMEType': 'video/x-flv',
            'Type': 'video',
            'Date': program.publish_date.strftime('%Y-%m-%dT%H:%M:%S')
            }
        if program.series is not None:
            self.details['Album'] = program.series.title
        if program.duration is not None:
            self.details['Duration'] = program.duration

    def get_details(self):
        return self.details
