import logging
import os
import sys

from twisted.internet import defer, interfaces, protocol, reactor
from twisted.web import resource, server
from zope.interface import implements


class RtmpdumpProcessProtocol(protocol.ProcessProtocol):
    """Publish output from rtmpdump to an HTTP connection."""
    implements(interfaces.IPushProducer)

    def __init__(self, request):
        self.request = request

    def connectionMade(self):
        self.request.registerProducer(self, True)
        self.transport.closeStdin()
        self.request.setHeader('content-type', 'video/x-flv')

    def errReceived(self, error):
        # Ignore error text.
        pass

    def outReceived(self, output):
        self.request.write(output)

    def processEnded(self, reason):
        if reason.value.exitCode != 0:
            logging.warning("rtmpdump exited with exit code %s" %
                            reason.value.exitCode)
        self.request.unregisterProducer()
        self.request.finish()

    def resumeProducing(self):
        """See `IPushProducer`."""
        self.transport.resumeProducing()

    def pauseProducing(self):
        """See `IPushProducer`."""
        self.transport.pauseProducing()

    def stopProducing(self):
        """See `IProducer`."""
        self.transport.loseConnection()


class ProgramResource(resource.Resource):

    isLeaf = True

    def __init__(self, guide, play_path):
        resource.Resource.__init__(self)
        self.guide = guide
        self.play_path = play_path

    def render_GET(self, request):
        d = self._stream_video(request)
        def errback(failure):
            request.finish()
            return failure
        d.addErrback(errback)
        return server.NOT_DONE_YET

    @defer.inlineCallbacks
    def _stream_video(self, request):
        # Get authentication details for the RTMP server.
        (server, token) = yield self.guide.get_auth_info()

        # Strip off .flv suffix from play path.
        play_path = self.play_path
        if play_path.endswith('.flv'):
            play_path = play_path[:-4]

        process_protocol = RtmpdumpProcessProtocol(request)
        rtmpdump_executable = os.environ.get('RTMPDUMP', 'rtmpdump')
        reactor.spawnProcess(
            process_protocol,
            os.environ.get('RTMPDUMP', 'rtmpdump'),
            ['rtmpdump',
             '--rtmp', server,
             '--auth', 'auth=%s' % token,
             '--tcUrl', '%s?auth=%s' % (server, token),
             '--playpath', play_path])


class ProgramCollection(resource.Resource):

    isLeaf = False

    def __init__(self, guide):
        resource.Resource.__init__(self)
        self.guide = guide

    def getChild(self, name, request):
        try:
            program_id = int(name)
        except ValueError:
            pass
        else:
            return ProgramResource(
                self.guide, self.guide.programs[program_id].video_url)
        return resource.Resource.getChild(self, name, request)


class Root(resource.Resource):

    isLeaf = False

    def __init__(self, guide):
        resource.Resource.__init__(self)
        self.guide = guide
        self.putChild('program', ProgramCollection(guide))

    def render_GET(self, request):
        return "Whatever"


def get_iview_site(guide):
    return server.Site(Root(guide))
